import { of as observableOf, ReplaySubject } from 'rxjs';
import { delay, tap, map, share } from 'rxjs/operators';
import { OcTableColumn } from './oc-table-column.model';
import { OcTableRow } from './oc-table-row.model';
import { OcTableConfigInitializer } from '../services/table-config-initializer';
import { OcTableConfigStoreService } from '../services/config-store.service';
import { has } from 'lodash-es';
export class OcTableDataSource {
    constructor(rows, columns, config) {
        // states & state change
        this.isLoading = false;
        this.pagerStateChangeInternal = new ReplaySubject(1);
        this.pagerStateChange = this.pagerStateChangeInternal.pipe(share());
        this.sorterStateChangeInternal = new ReplaySubject(1);
        this.sorterStateChange = this.sorterStateChangeInternal.pipe(share());
        // internal
        this.pagerState = {};
        this.sorterState = null; // array of {field: field, dir: dir}
        this.rows = rows;
        this.columns = columns;
        this.renewCellsRef();
        if (!!config) {
            this.config = config;
        }
    }
    static constructFromData(data) {
        let source = new OcTableDataSource([], []);
        source.data = data;
        source.reloadFromData(data);
        return source;
    }
    static constructFromDataAndConfig(data, config) {
        if (!config) {
            return this.constructFromData(data);
        }
        // create
        let dataSource = new OcTableDataSource([], [], config);
        dataSource.data = data;
        // init
        dataSource.initFromConfig(config);
        dataSource.reloadFromData(data);
        const sortedColumn = dataSource.getSortedColumn();
        // default sorter and default page
        if (config.pager && config.pager.perPage && sortedColumn) {
            dataSource.setPagerConf(config.pager.defaultPage || 1, config.pager.perPage);
            dataSource.setSortConf(sortedColumn.config.name, sortedColumn.config.sortOrder);
            sortedColumn.datasource.sort(sortedColumn.config.name, sortedColumn.config.sortOrder, null, config.pager.defaultPage).subscribe();
        }
        else if (config.pager && config.pager.perPage && !sortedColumn) {
            dataSource.setPagerConf(config.pager.defaultPage || 1, config.pager.perPage);
            dataSource.paginate(config.pager.defaultPage || 1, config.pager.perPage).subscribe();
        }
        return dataSource;
    }
    /**
     * Reload DataSource,
     * @param {Array<any>} data
     */
    reloadFromData(data) {
        // pre set data by config
        let configInitializer = new OcTableConfigInitializer();
        this.config = configInitializer.mergeRuntimeConfig(this.config);
        configInitializer.validateConfig(data, this.config);
        data = configInitializer.filterData(data, this.config);
        // create columns, only when first reload
        if ((!this.columns || !this.columns.length) && (!this.config || !this.config.columns)) {
            this.columns = this.createColumnsFromData(data);
        }
        if ((!this.columns || !this.columns.length) && !!this.config && !!this.config.columns) {
            this.columns = this.createColumnsFromConfig(this.config);
        }
        // create rows
        this.rows = data.map(d => OcTableRow.constructRowFromData(d));
        // renew ref
        this.renewCellsRef();
        this.renewConfigRef();
        configInitializer.bindPipes(this, this.config);
        // this.emitOnChanged({event: 'reload'})
    }
    /**
     * Accessors
     * @param {string} name
     * @returns {OcTableColumn}
     */
    getColumnByName(name) {
        return this.columns.find(el => {
            return (el.value === name);
        });
    }
    /**
     * ChangeSource
     */
    /**
     * TableDataOperation Interface
     */
    setSortConf(field, dir) {
        let index = this.columns.findIndex(col => col.value === field);
        if (index !== -1) {
            // clear others direction first
            this.columns.forEach(col => col.sortOrder = null);
            // set this sort conf
            this.columns[index].sortOrder = dir;
        }
        this.sorterStateChangeInternal.next({
            field: field,
            dir: dir
        });
    }
    getSortConf() {
        let col = this.columns.find(col => col.sortOrder !== null);
        return col ? { field: col.value, dir: col.sortOrder } : null;
    }
    getPagerConf() {
        return this.pagerState;
    }
    getSortedColumn() {
        return this.columns.find((col) => !!col.config.sortOrder && col.config.sortable);
    }
    setPagerConf(curPage, perPage, totalRows) {
        this.pagerState['curPage'] = curPage;
        this.pagerState['perPage'] = perPage;
        this.pagerState['totalRows'] = totalRows;
        this.pagerStateChangeInternal.next(this.pagerState);
        return this.pagerState;
    }
    sort(field, dir, compareFunc, defaultPage) {
        // set isLoading
        this.isLoading = true;
        if (!this.config.asyncDataFetchFn) {
            // default compare function
            let COMPARE = (direction, a, b) => {
                if (a < b) {
                    return -1 * direction;
                }
                if (a > b) {
                    return direction;
                }
                return 0;
            };
            // sort
            let direct = (dir === 'asc') ? 1 : -1;
            let compare = compareFunc ? compareFunc : COMPARE;
            this.data.sort((a, b) => {
                return compare.call(null, direct, a[field], b[field]);
            });
            // setConf
            this.setSortConf(field, dir);
            // reload
            if (!!this.pagerState && !!this.pagerState['perPage']) {
                return this.paginate((defaultPage || 1), this.pagerState['perPage']).pipe(tap(res => this.isLoading = false));
            }
            else {
                this.reloadFromData(this.data);
                return observableOf(this).pipe(delay(500), tap(res => this.isLoading = false));
            }
            // server data
        }
        else {
            this.setPagerConf((defaultPage || 1), this.getPagerConf()['perPage']);
            return this.config.asyncDataFetchFn(this.getPagerConf(), this.getSortConf()).pipe(map(res => {
                this.setPagerConf(this.getPagerConf()['curPage'], this.getPagerConf()['perPage'], res.totalRows);
                this.reloadFromData(res.data);
                return this;
            }), delay(50), tap(res => this.isLoading = false));
        }
    }
    paginate(page, perPage) {
        this.isLoading = true;
        // local data;
        if (!this.config.asyncDataFetchFn) {
            let newData = this.data.slice(perPage * (page - 1), perPage * page);
            this.reloadFromData(newData);
            this.setPagerConf(page, perPage, this.data.length);
            return observableOf(this).pipe(delay(500), tap(res => this.isLoading = false));
            // server data
        }
        else {
            this.setPagerConf(page, perPage);
            return this.config.asyncDataFetchFn(this.getPagerConf(), this.getSortConf()).pipe(map(res => {
                this.setPagerConf(page, perPage, res.totalRows);
                this.reloadFromData(res.data);
                return this;
            }), delay(50), tap(res => this.isLoading = false));
        }
    }
    /**
     * Other Operation
     */
    reorderColumn(colNameWithOrder) {
        if (!colNameWithOrder || !colNameWithOrder.length ||
            !colNameWithOrder.every(s => this.columns.findIndex(col => col.value === s) !== -1)) {
            throw new Error('Invalid Table Order');
        }
        // reorder colum
        let replaceLen = this.columns.length;
        colNameWithOrder.forEach(x => {
            let col = this.columns.find(col => col.value === x);
            if (!col)
                throw new Error('Invalid Table Order: cannot find column with column name:' + x);
            this.columns.push(col);
        });
        this.columns.splice(0, replaceLen);
        // reorder cell
        this.rows.forEach(row => {
            let replaceLen = row.cells.length;
            colNameWithOrder.forEach(x => {
                let cell = row.cells.find(cell => cell.column.value === x);
                if (!cell)
                    throw new Error('Invalid Table Order: cannot find cell with column name:' + x);
                row.cells.push(cell);
            });
            row.cells.splice(0, replaceLen);
        });
        // save current runtimeConfig to saveConfig
        let rtConfigStore = new OcTableConfigStoreService();
        rtConfigStore.saveConfigByDataSource(this);
    }
    /**
     * @Internal
     */
    initFromConfig(config) {
        if (has(config, 'name'))
            this.translatePipePrefix = config.name;
        if (has(config, 'dropdown'))
            this.dropdown = config.dropdown;
        if (has(config, 'floatingButtons'))
            this.floatingButtons = config.floatingButtons;
    }
    createColumnsFromData(data) {
        // extract column keys form data
        let columnsKeys = data.reduce((keys, el) => {
            // filter out the redundant keys
            let filteredKeys = [...Array.from(el.keys())].filter(key => keys.indexOf(key) === -1);
            return keys.concat(filteredKeys);
        }, []);
        // create columns if need, why do this
        let ocColumns = columnsKeys.map(obj => {
            if (this.getColumnByName(obj)) {
                // only change value if exits, TODO: question about why do this?
                this.getColumnByName(obj).value = obj;
                return this.getColumnByName(obj);
            }
            else {
                return new OcTableColumn(obj);
            }
        });
        return ocColumns;
    }
    createColumnsFromConfig(config) {
        if (!config || !config.columns || !config.columns.length)
            return;
        return config.columns.map(x => new OcTableColumn(x.name));
    }
    renewCellsRef() {
        if (!this.rows || !this.columns || this.columns.length === 0)
            return;
        // cells
        this.rows.forEach((row) => {
            row.cells.forEach((cell) => {
                cell.row = row;
                let index = row.cells.indexOf(cell);
                cell.column = index !== -1 ? this.columns[index] : null;
                cell.datasource = this;
            });
        });
        // columns
        this.columns.forEach(col => col.datasource = this);
        return;
    }
    renewConfigRef() {
        if (!this.config || !this.config.columns || !this.config.columns.length)
            return;
        this.config.columns.forEach(x => {
            let col = this.getColumnByName(x.name);
            if (!!col)
                col.config = x;
        });
    }
}
